# Copyright 2018, 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
from __future__ import division

import socket
import time
import json
import select
from netq_lib.common.enums import ConfKw
from netq_agent.cmd.netq_agent_discovery import NetQAgentDiscovery, AGENTSTATUS, ChannelState

from concurrent import futures

# GRPC Specfic Imports
import grpc
from netq_agent.interface.netq_pb2 import (
    NetqAgentMessage,
)
from netq_agent.interface.netq_pb2_grpc import (
    NetqAgentStreamServiceServicer,
    add_NetqAgentStreamServiceServicer_to_server
)
PARTIAL_TRANSFORM = 1

class NetQAgentServicer(NetqAgentStreamServiceServicer):

    def __init__(self, logger, pq):
        self._logger = logger
        self.seq_no = 0
        self._pq = pq

    def service_start_request(self):
        self._start = True

    def get_lldp_command_out(self):
        msg = ""
        return msg

    def get_sequence_number(self):
        self.seq_no = self.seq_no + 1
        return self.seq_no

    def message_builder(self, msg_type, json_out):

        netq_message = NetqAgentMessage(aid=socket.gethostname(),
                                        message_type=msg_type,
                                        seqno=self.get_sequence_number(),
                                        ts=long(time.time() * 1000),
                                        trans_mode=PARTIAL_TRANSFORM,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def StreamMessage(self, request, context):
        # pq = NetQPriorityQueue.PQ  # Wrap as utility function
        pq = self._pq
        fd_set = pq.get_reader_fds()
        self._logger.info("Enable GRPC Priority Queue len %d", pq.len(1))
        self._logger.debug("Priority Queue Size {}".format(fd_set))
        while True:
            try:
                (epoll_input, [], []) = select.select(fd_set, [], [])
                for queue_idx in range(0, self._pq.max_num_queue):
                    msg_dict_array = {}
                    for priority, message_type, msg in pq.get_buffer(queue_idx):
                        #json_obj = json.loads(json.dumps(
                        #    msg.data).decode('string-escape').strip('"'))
                        #message_type = json_obj["messageType"]
                        if message_type not in msg_dict_array:
                            msg_dict_array[message_type] = []
                        if isinstance(msg.data, list):
                            msg_dict_array[message_type] = msg.data
                        else:
                            msg_dict_array[message_type].append(msg.data)
                    for message_type, msg_array in msg_dict_array.iteritems():
                        rpc_msg = self.message_builder(message_type, msg_array)
                        yield rpc_msg
            except Exception as e:
                self._logger.error("StreamMessage failed {}".format(e))
                return


class NetQAgentStream(object):

    def __init__(self, conf, logger, pq):
        self._conf = conf
        self._channel = None
        self._logger = logger
        self._pq = pq
        self.discovery_agent = None

    def _get_server_addr(self):
        # always listen on *:port
        agent_url = "0.0.0.0"
        try:
            agent_port = self._conf['netq-agent'].get(ConfKw.AGENT_PORT, ConfKw.DEFAULT_AGENT_PORT)
        except Exception:
            agent_port = ConfKw.DEFAULT_AGENT_PORT
        server_addr = "{}:{}".format(agent_url, agent_port)
        return server_addr

    def _init_grpc_server(self):
        if getattr(self, 'server', None) is not None:
            self.server.stop(None)
        self.server = None
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
        add_NetqAgentStreamServiceServicer_to_server(
            NetQAgentServicer(self._logger, self._pq), self.server)

        server_addr = self._get_server_addr()
        self._logger.info("Start Streaming server at {}".format(server_addr))

        self.server.add_insecure_port(server_addr)
        self.server.start()
        self._logger.info("Streaming Server Started")


    def _init_rpc_service(self):

        self._init_grpc_server()

        self.discovery_agent = NetQAgentDiscovery(
            self._conf, logger=self._logger)

        if self.discovery_agent is None:
            self._logger.error("Failed to start disovery process")
            self.server.stop(None)
            return AGENTSTATUS.FAILURE

        discovery_enabled = self._is_discovery_enabled()
        self.wait_time = 10
        self.discovery_retry_time = 60
        retry_count = 0
        prev_wait_time = time.time()
        period_cnt = 0
        while True:
            curr_time = time.time()
            if discovery_enabled:
                if self.get_discovery_status() == ChannelState.IDLE:
                    if not retry_count:
                        self._logger.info("OPTA Start Discovery IDLE")
                    elif retry_count < 3:
                        self._logger.info("OPTA Re-Start Discovery (retry %s) IDLE", retry_count)
                    self.discovery_agent.close()
                    ret = self.discovery_agent.start_discovery()
                    if ret == AGENTSTATUS.FAILURE:
                        # big hammer - start/stop the GRPC server
                        self._logger.info("OPTA Discovery failure - stop/start streaming server")
                        self._init_grpc_server()
                        self._logger.info("OPTA Restarted GRPC server")
                    retry_count += 1
                    # reset the periodic discovery logic
                    prev_wait_time = curr_time
                elif self.get_discovery_status() == ChannelState.CONNECTED:
                    if retry_count:
                        self._logger.info("OPTA Discovery OK READY")
                        retry_count = 0
                    if (curr_time - prev_wait_time) >= self.discovery_retry_time:
                        # calculate next retry time
                        prev_wait_time = curr_time
                        if period_cnt == 10:
                            self._logger.info("OPTA Periodic Discovery")
                            period_cnt = 0
                        period_cnt += 1
                        self.discovery_agent.close()
                        self.discovery_agent.start_discovery()
                else:
                    # reset the periodic discovery logic
                    prev_wait_time = curr_time
                    if not retry_count:
                        self._logger.info("OPTA Discovery Channel not ready")
                        retry_count += 1
            time.sleep(self.wait_time)

    def start_stream_server(self):
        pass

    def start_stream(self):
        pass

    def stop_stream(self):
        pass

    def get_discovery_status(self):
        return self.discovery_agent.channel_state

    def _is_discovery_enabled(self):
        try:
            bool_obj = self._conf['netq-agent'].get(ConfKw.ENABLE_OPTA_DISCOVERY, ConfKw.DEFAULT_ENABLE_OPTA_DISCOVERY)
            if isinstance(bool_obj, bool):
                if bool_obj:
                    return True
                else:
                    return False
        except Exception:
            return True

        if bool_obj.lower() in ['yes', 'true', '1']:
            return True
        else:
            return False

    def stream_main(self):

        self._init_rpc_service()
        return
